@@ -56,7 +56,7 @@ enum
 enum
 {
 	PROP_0,
-	PROP_HAS_UNSAVED_CHANGES,
+	PROP_MODIFIED,
 	PROP_HAS_SELECTION,
 	PROP_PATH,
 	PROP_READ_ONLY
@@ -64,21 +64,21 @@ enum
 
 struct _GladeProjectPrivate
 {
-	gchar *path;     /* The full canonical path of the glade file for this project */
+	gchar *path;            /* The full canonical path of the glade file for this project */
 
 	guint   instance_count; /* How many projects with this name */
 
-	gint   unsaved_number; /* A unique number for this project if it is untitled */
+	gint   unsaved_number;  /* A unique number for this project if it is untitled */
 
-	gboolean readonly; /* A flag that is set if the project is readonly */
+	gboolean readonly;      /* A flag that is set if the project is readonly */
 
-	gboolean loading;/* A flags that is set when the project is loading */
+	gboolean loading;       /* A flags that is set when the project is loading */
 	
-	gboolean changed;    /* A flag that is set when a project has changes
-			      * if this flag is not set we don't have to query
-			      * for confirmation after a close or exit is
-			      * requested
-			      */
+	gboolean modified;    /* A flag that is set when a project has unsaved modifications
+			       * if this flag is not set we don't have to query
+			       * for confirmation after a close or exit is
+			       * requested
+			       */
 
 	GList *objects; /* A list of #GObjects that make up this project.
 			 * The objects are stored in no particular order.
@@ -91,23 +91,28 @@ struct _GladeProjectPrivate
 			   * of #GtkWidget items.
 			   */
 
-	gboolean has_selection; /* Whether the project has a selection */
+	gboolean     has_selection;           /* Whether the project has a selection */
 
-	GList *undo_stack; /* A stack with the last executed commands */
-	GList *prev_redo_item; /* Points to the item previous to the redo items */
-	GHashTable *widget_names_allocator; /* hash table with the used widget names */
-	GHashTable *widget_old_names; /* widget -> old name of the widget */
+	GList       *undo_stack;              /* A stack with the last executed commands */
+	GList       *prev_redo_item;          /* Points to the item previous to the redo items */
+	GHashTable  *widget_names_allocator;  /* hash table with the used widget names */
+	GHashTable  *widget_old_names;        /* widget -> old name of the widget */
 	GtkTooltips *tooltips;
 	
+	GladeCommand *first_modification; /* we record the first modification, so that we
+	                                   * can set "modification" to FALSE when we
+	                                   * undo this modification
+	                                   */
+	
 	GtkAccelGroup *accel_group;
 
 	GHashTable *resources; /* resource filenames & thier associated properties */
 	
-	gchar *comment; /* XML comment, Glade will preserve whatever comment was
-			 * in file, so users can delete or change it.
-			 */
+	gchar *comment;        /* XML comment, Glade will preserve whatever comment was
+			        * in file, so users can delete or change it.
+			        */
 			 
-	time_t  mtime; /* last UTC modification time of file, or 0 if it could not be read */
+	time_t  mtime;         /* last UTC modification time of file, or 0 if it could not be read */
 };
 
 
@@ -222,8 +227,8 @@ glade_project_get_property (GObject    *object,
 
 	switch (prop_id)
 	{
-		case PROP_HAS_UNSAVED_CHANGES:
-			g_value_set_boolean (value, project->priv->changed);
+		case PROP_MODIFIED:
+			g_value_set_boolean (value, project->priv->modified);
 			break;
 		case PROP_HAS_SELECTION:
 			g_value_set_boolean (value, project->priv->has_selection);
@@ -240,6 +245,48 @@ glade_project_get_property (GObject    *object,
 	}
 }
 
+/**
+ * glade_project_set_modified:
+ * @project: a #GladeProject
+ * @modified: Whether the project should be set as modified or not
+ * @modification: The first #GladeCommand which caused the project to have unsaved changes
+ *
+ * Set's whether a #GladeProject should be flagged as modified or not. This is useful
+ * for indicating that a project has unsaved changes. If @modified is #TRUE, then
+ * @modification will be recorded as the first change which caused the project to 
+ * have unsaved changes. @modified is #FALSE then @modification will be ignored.
+ *
+ * If @project is already flagged as modified, then calling this method with
+ * @modified as #TRUE, will have no effect. Likewise, if @project is unmodified
+ * then calling this method with @modified as #FALSE, will have no effect.
+ *
+ */
+static void
+glade_project_set_modified (GladeProject *project,
+			    gboolean      modified,
+			    GladeCommand *modification)
+{
+	GladeProjectPrivate *priv = project->priv;
+
+	if (priv->modified != modified)
+	{
+		priv->modified = !priv->modified;
+		
+		if (priv->modified)
+		{
+			g_assert (priv->first_modification == NULL);
+			g_assert (modification != NULL);
+			priv->first_modification = modification;
+		}
+		else
+		{
+			g_assert (priv->first_modification != NULL);
+			priv->first_modification = NULL;
+		}
+		
+		g_object_notify (G_OBJECT (project), "modified");
+	}
+}
 
 /*******************************************************************
                           GladeProjectClass
@@ -266,7 +313,7 @@ glade_project_undo_impl (GladeProject *project)
 	GladeCommand *cmd, *next_cmd;
 
 	while ((cmd = glade_project_next_undo_item (project)) != NULL)
-	{
+	{	
 		glade_command_undo (cmd);
 
 		glade_project_walk_back (project);
@@ -328,13 +375,14 @@ glade_project_next_redo_item_impl (GladeProject *project)
 static void
 glade_project_push_undo_impl (GladeProject *project, GladeCommand *cmd)
 {
-	GList* tmp_redo_item;
+	GladeProjectPrivate *priv = project->priv;
+	GList *tmp_redo_item;
 
 	/* If there are no "redo" items, and the last "undo" item unifies with
 	   us, then we collapse the two items in one and we're done */
-	if (project->priv->prev_redo_item != NULL && project->priv->prev_redo_item->next == NULL)
+	if (priv->prev_redo_item != NULL && priv->prev_redo_item->next == NULL)
 	{
-		GladeCommand *cmd1 = project->priv->prev_redo_item->data;
+		GladeCommand *cmd1 = priv->prev_redo_item->data;
 		
 		if (glade_command_unifies (cmd1, cmd))
 		{
@@ -349,32 +397,38 @@ glade_project_push_undo_impl (GladeProject *project, GladeCommand *cmd)
 	}
 
 	/* We should now free all the "redo" items */
-	tmp_redo_item = g_list_next (project->priv->prev_redo_item);
+	tmp_redo_item = g_list_next (priv->prev_redo_item);
 	while (tmp_redo_item)
 	{
 		g_assert (tmp_redo_item->data);
+		
+		/* just for safety, we might not need this */
+		if (GLADE_COMMAND (tmp_redo_item->data) == priv->first_modification)
+			priv->first_modification = NULL;
+		
 		g_object_unref (G_OBJECT (tmp_redo_item->data));
+		
 		tmp_redo_item = g_list_next (tmp_redo_item);
 	}
 
-	if (project->priv->prev_redo_item)
+	if (priv->prev_redo_item)
 	{
-		g_list_free (g_list_next (project->priv->prev_redo_item));
-		project->priv->prev_redo_item->next = NULL;
+		g_list_free (g_list_next (priv->prev_redo_item));
+		priv->prev_redo_item->next = NULL;
 	}
 	else
 	{
-		g_list_free (project->priv->undo_stack);
-		project->priv->undo_stack = NULL;
+		g_list_free (priv->undo_stack);
+		priv->undo_stack = NULL;
 	}
 
 	/* and then push the new undo item */
-	project->priv->undo_stack = g_list_append (project->priv->undo_stack, cmd);
+	priv->undo_stack = g_list_append (priv->undo_stack, cmd);
 
 	if (project->priv->prev_redo_item == NULL)
-		project->priv->prev_redo_item = project->priv->undo_stack;
+		priv->prev_redo_item = priv->undo_stack;
 	else
-		project->priv->prev_redo_item = g_list_next (project->priv->prev_redo_item);
+		priv->prev_redo_item = g_list_next (priv->prev_redo_item);
 
 
 	g_signal_emit (G_OBJECT (project),
@@ -387,12 +441,16 @@ glade_project_changed_impl (GladeProject *project,
 			    GladeCommand *command,
 			    gboolean      forward)
 {
-	if (!project->priv->changed && !project->priv->loading)
+	if (!project->priv->loading)
 	{
-		project->priv->changed = TRUE;
-		g_object_notify (G_OBJECT (project), "has-unsaved-changes");
+		/* if this command is the first modification to cause the project
+		 * to have unsaved changes, then we can now flag the project as unmodified
+		 */
+		if (command == project->priv->first_modification)
+			glade_project_set_modified (project, FALSE, NULL);
+		else
+			glade_project_set_modified (project, TRUE, command);				
 	}
-
 	glade_app_update_ui ();
 }
 
@@ -414,6 +472,7 @@ glade_project_init (GladeProject *project)
 	priv->has_selection = FALSE;
 	priv->undo_stack = NULL;
 	priv->prev_redo_item = NULL;
+	priv->first_modification = NULL;
 	priv->widget_names_allocator = g_hash_table_new_full (g_str_hash,
 							      g_str_equal,
 							      g_free, 
@@ -428,7 +487,7 @@ glade_project_init (GladeProject *project)
 						 g_direct_equal, 
 						 NULL, g_free);
 
-	priv->unsaved_number = glade_id_allocator_allocate (get_unsaved_number_allocator ());
+	priv->unsaved_number = glade_id_allocator_allocate (get_unsaved_number_allocator ());	
 }
 
 static void
@@ -622,10 +681,10 @@ glade_project_class_init (GladeProjectClass *klass)
 			      0);
 
 	g_object_class_install_property (object_class,
-					 PROP_HAS_UNSAVED_CHANGES,
-					 g_param_spec_boolean ("has-unsaved-changes",
-							       _("Has Unsaved Changes"),
-							       _("Whether project has unsaved changes"),
+					 PROP_MODIFIED,
+					 g_param_spec_boolean ("modified",
+							       _(""),
+							       _("Whether project has beem modified since it was last saved"),
 							       FALSE,
 							       G_PARAM_READABLE));
 
@@ -728,7 +787,7 @@ glade_project_load_from_file (GladeProject *project, const gchar *path)
 	if (glade_util_file_is_writeable (project->priv->path) == FALSE)
 		glade_project_set_readonly (project, TRUE);
 
-	project->priv->changed = FALSE;
+	project->priv->modified = FALSE;
 		
 	project->priv->mtime = glade_util_get_file_mtime (project->priv->path, NULL);
 
@@ -881,6 +940,10 @@ glade_project_add_object (GladeProject *project,
 	if ((gwidget = glade_widget_get_from_gobject (object)) == NULL)
 		return;
 
+	/* Dont add widgets that are already in the project */
+	if (glade_project_has_object (project, object))
+		return;
+		
 	/* Code body starts here */
 	reentrancy_count++;
 
@@ -942,7 +1005,7 @@ glade_project_has_object (GladeProject *project, GObject *object)
  *
  * TODO: Write me
  */
-void
+static void
 glade_project_release_widget_name (GladeProject *project, GladeWidget *glade_widget, const char *widget_name)
 {
 	const char *first_number = widget_name;
@@ -1160,7 +1223,7 @@ glade_project_new_widget_name (GladeProject *project, const char *base_name)
 	return name;
 }
 
-void
+static void
 glade_project_set_has_selection (GladeProject *project, gboolean has_selection)
 {
 	g_assert (GLADE_IS_PROJECT (project));
@@ -1378,8 +1441,8 @@ glade_project_make_comment ()
 {
 	time_t now = time (NULL);
 	gchar *comment;
-	comment = g_strdup_printf (GLADE_XML_COMMENT" "PACKAGE_VERSION" on %sby %s@%s",
-				   ctime (&now), g_get_user_name (), g_get_host_name ());
+	comment = g_strdup_printf (GLADE_XML_COMMENT" "PACKAGE_VERSION" on %s",
+				   ctime (&now));
 	glade_util_replace (comment, '\n', ' ');
 	
 	return comment;
@@ -1590,7 +1653,7 @@ glade_project_load_from_interface (GladeProject   *project,
 	/* Reset project status here too so that you get a clean
 	 * slate after calling glade_project_open().
 	 */
-	project->priv->changed = FALSE;
+	project->priv->modified = FALSE;
 	project->priv->loading = FALSE;
 
 	/* Emit "parse-finished" signal */
@@ -1722,12 +1785,8 @@ glade_project_save (GladeProject *project, const gchar *path, GError **error)
 				    !glade_util_file_is_writeable (project->priv->path));
 
 	project->priv->mtime = glade_util_get_file_mtime (project->priv->path, NULL);
-
-	if (project->priv->changed)
-	{
-		project->priv->changed = FALSE;
-		g_object_notify (G_OBJECT (project), "has-unsaved-changes");
-	}
+	
+	glade_project_set_modified (project, FALSE, NULL);
 
 	if (project->priv->unsaved_number > 0)
 	{
@@ -1816,6 +1875,151 @@ glade_project_push_undo (GladeProject *project, GladeCommand *cmd)
 	GLADE_PROJECT_GET_CLASS (project)->push_undo (project, cmd);
 }
 
+static GList *
+walk_command (GList *list, gboolean forward)
+{
+	GladeCommand *cmd = list->data;
+	GladeCommand *next_cmd;
+
+	if (forward)
+		list = list->next;
+	else
+		list = list->prev;
+	
+	next_cmd = list ? list->data : NULL;
+	
+	while (list && next_cmd->group_id != 0 &&
+	       next_cmd->group_id == cmd->group_id)
+	{
+		if (forward)
+			list = list->next;
+		else
+			list = list->prev;
+
+		if (list)
+			next_cmd = list->data;
+	}
+
+	return list;
+}
+
+static void
+undo_item_activated (GtkMenuItem  *item,
+		     GladeProject *project)
+{
+	gint index, next_index;
+	
+	GladeCommand *cmd = g_object_get_data (G_OBJECT (item), "command-data");
+	GladeCommand *next_cmd;
+
+	index = g_list_index (project->priv->undo_stack, cmd);
+
+	do
+	{
+		next_cmd = glade_project_next_undo_item (project);
+		next_index = g_list_index (project->priv->undo_stack, next_cmd);
+
+		glade_project_undo (project);
+		
+	} while (next_index > index);
+}
+
+static void
+redo_item_activated (GtkMenuItem  *item,
+		     GladeProject *project)
+{
+	gint index, next_index;
+	
+	GladeCommand *cmd = g_object_get_data (G_OBJECT (item), "command-data");
+	GladeCommand *next_cmd;
+
+	index = g_list_index (project->priv->undo_stack, cmd);
+
+	do
+	{
+		next_cmd = glade_project_next_redo_item (project);
+		next_index = g_list_index (project->priv->undo_stack, next_cmd);
+
+		glade_project_redo (project);
+		
+	} while (next_index < index);
+}
+
+
+/**
+ * glade_project_undo_items:
+ * @project: A #GladeProject
+ *
+ * Creates a menu of the undo items in the project stack
+ *
+ * Returns: A newly created menu
+ */
+GtkWidget *
+glade_project_undo_items (GladeProject *project)
+{
+	GtkWidget *menu = NULL;
+	GtkWidget *item;
+	GladeCommand *cmd;
+	GList   *l;
+
+	for (l = project->priv->prev_redo_item; l; l = walk_command (l, FALSE))
+	{
+		cmd = l->data;
+
+		if (!menu) menu = gtk_menu_new ();
+		
+		item = gtk_menu_item_new_with_label (cmd->description);
+		gtk_widget_show (item);
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), GTK_WIDGET (item));
+		g_object_set_data (G_OBJECT (item), "command-data", cmd);
+
+		g_signal_connect (G_OBJECT (item), "activate",
+				  G_CALLBACK (undo_item_activated), project);
+
+	}
+	
+	return menu;
+}
+
+/**
+ * glade_project_redo_items:
+ * @project: A #GladeProject
+ *
+ * Creates a menu of the undo items in the project stack
+ *
+ * Returns: A newly created menu
+ */
+GtkWidget *
+glade_project_redo_items (GladeProject *project)
+{
+	GtkWidget *menu = NULL;
+	GtkWidget *item;
+	GladeCommand *cmd;
+	GList   *l;
+
+	for (l = project->priv->prev_redo_item ?
+		     project->priv->prev_redo_item->next :
+		     project->priv->undo_stack;
+	     l; l = walk_command (l, TRUE))
+	{
+		cmd = l->data;
+
+		if (!menu) menu = gtk_menu_new ();
+		
+		item = gtk_menu_item_new_with_label (cmd->description);
+		gtk_widget_show (item);
+		gtk_menu_shell_append (GTK_MENU_SHELL (menu), GTK_WIDGET (item));
+		g_object_set_data (G_OBJECT (item), "command-data", cmd);
+
+		g_signal_connect (G_OBJECT (item), "activate",
+				  G_CALLBACK (redo_item_activated), project);
+
+	}
+	
+	return menu;
+}
+
+
 void
 glade_project_reset_path (GladeProject *project)
 {
@@ -2048,7 +2252,7 @@ glade_project_get_name (GladeProject *project)
  * glade_project_is_loading:
  * @project: A #GladeProject
  *
- * Returns: Whether the project is being loaded or not.
+ * Returns: Whether the project is being loaded or not
  *       
  */
 gboolean
@@ -2091,11 +2295,19 @@ glade_project_set_instance_count (GladeProject *project, guint instance_count)
 	project->priv->instance_count = instance_count;
 }
 
+/** 
+ * glade_project_get_modified:
+ * @project: a #GladeProject
+ *
+ * Get's whether the project has been modified since it was last saved.
+ *
+ * Returns: #TRUE if the project has been modified since it was last saved
+ */ 
 gboolean
-glade_project_get_has_unsaved_changes (GladeProject *project)
+glade_project_get_modified (GladeProject *project)
 {
 	g_return_val_if_fail (GLADE_IS_PROJECT (project), FALSE);
 
-	return project->priv->changed;
+	return project->priv->modified;
 }
 
